@@ -3,12 +3,12 @@
 import asyncio
 import enum
 import io
-import os
-import warnings
 from logging import Logger
-from typing import Optional
+from pathlib import Path
+from typing import Any, Optional, Union
 
 from shellous.log import LOG_DETAIL, log_method
+from shellous.pty_util import PtyAdapterOrBool
 from shellous.util import decode
 
 _CHUNK_SIZE = 8192
@@ -24,6 +24,7 @@ class Redirect(enum.IntEnum):
     DEVNULL = asyncio.subprocess.DEVNULL  # -3
     CAPTURE = -10
     INHERIT = -11
+    RESULT = -12
     DEFAULT = -20
 
     def is_custom(self):
@@ -31,11 +32,12 @@ class Redirect(enum.IntEnum):
         return self in {
             Redirect.CAPTURE,
             Redirect.INHERIT,
+            Redirect.RESULT,
             Redirect.DEFAULT,
         }
 
     @staticmethod
-    def from_default(obj, fdesc, pty):
+    def from_default(obj: Any, fdesc: int, pty: PtyAdapterOrBool):
         "Return object with Redirect.DEFAULT replaced by actual value."
         if not isinstance(obj, Redirect) or obj != Redirect.DEFAULT:
             return obj
@@ -44,23 +46,16 @@ class Redirect(enum.IntEnum):
         return _DEFAULT_REDIRECTION[(fdesc, bool(pty))]
 
     @staticmethod
-    def from_literal(literal, is_stderr=False):
+    def from_literal(literal: Any, is_stderr: bool = False):
         "Return object with literal values replaced by Redirect constant."
 
         if isinstance(literal, (tuple, type(...), type(None))):
-            warnings.warn(
-                f"literal {literal!r} is no longer supported",
-                DeprecationWarning,
-            )
+            raise TypeError(f"literal {literal!r} is unsupported")
 
         # For stderr, the literal `1` indicates that stderr is redirected to
         # the same place as STDOUT.
         if is_stderr and isinstance(literal, int) and literal == 1:
-            warnings.warn(
-                f"stderr literal {literal!r} is no longer supported",
-                DeprecationWarning,
-            )
-            return Redirect.STDOUT
+            raise TypeError(f"stderr literal {literal!r} is unupported")
 
         try:
             return _LITERAL_REDIRECT.get(literal, literal)
@@ -70,17 +65,17 @@ class Redirect(enum.IntEnum):
 
 # This table has the default redirections for (src, pty).
 # Sources are stdin, stdout, stderr. Used by Redirect.from_default().
-_DEFAULT_REDIRECTION = {
+_DEFAULT_REDIRECTION: dict[tuple[int, bool], Union[bytes, Redirect]] = {
     # (FD, PTY)
     (_STDIN, False): b"",
     (_STDIN, True): Redirect.CAPTURE,
     (_STDOUT, False): Redirect.CAPTURE,
     (_STDOUT, True): Redirect.CAPTURE,
-    (_STDERR, False): Redirect.DEVNULL,
+    (_STDERR, False): Redirect.RESULT,
     (_STDERR, True): Redirect.STDOUT,
 }
 
-_LITERAL_REDIRECT = {
+_LITERAL_REDIRECT: dict[Any, Redirect] = {
     None: Redirect.DEVNULL,
     ...: Redirect.INHERIT,
     (): Redirect.CAPTURE,
@@ -90,33 +85,25 @@ _LITERAL_REDIRECT = {
 STDIN_TYPES = (
     str,
     bytes,
-    os.PathLike,
+    Path,
     bytearray,
     io.IOBase,
     int,
     Redirect,
     asyncio.StreamReader,
-    type(None),
-    type(...),
-    tuple,
 )
 
 STDOUT_TYPES = (
-    str,
-    bytes,
-    os.PathLike,
+    Path,
     bytearray,
     io.IOBase,
     int,
     Redirect,
     Logger,
     asyncio.StreamWriter,
-    type(None),
-    type(...),
-    tuple,
 )
 
-STDOUT_APPEND_TYPES = (str, bytes, os.PathLike)
+STDOUT_APPEND_TYPES = (Path,)
 
 
 async def _drain(stream: asyncio.StreamWriter):
@@ -250,6 +237,29 @@ async def copy_bytearray(source: asyncio.StreamReader, dest: bytearray):
         dest.extend(data)
 
 
+@log_method(LOG_DETAIL)
+async def copy_bytearray_limit(
+    source: asyncio.StreamReader,
+    dest: bytearray,
+    limit: int,
+):
+    "Copy limited number of bytes from source stream to dest bytearray."
+    while True:
+        data = await source.read(_CHUNK_SIZE)
+        if not data:
+            break
+        dest.extend(data)
+        if len(dest) > limit:
+            del dest[limit:]
+            break
+
+    if data:
+        # After reaching the limit, continue to read and discard bytes from
+        # source to avoid possible blocking/deadlock inside the source program.
+        while await source.read(_CHUNK_SIZE):
+            pass
+
+
 @log_method(LOG_DETAIL)
 async def copy_streamwriter(source: asyncio.StreamReader, dest: asyncio.StreamWriter):
     "Copy bytes from source stream to dest StreamWriter."
