@@ -1,11 +1,16 @@
 "Implements support for Results."
 
 import asyncio
+import sys
 from dataclasses import dataclass
-from typing import Any, Optional
+from typing import Any, Optional, Union
 
+import shellous
 from shellous.util import decode
 
+# Limit number of bytes of stderr stored in Result object.
+RESULT_STDERR_LIMIT = 1024
+
 
 class ResultError(Exception):
     "Represents a non-zero exit status."
@@ -16,15 +21,21 @@ class ResultError(Exception):
         return self.args[0]
 
 
-@dataclass(frozen=True)
+_KW_ONLY = {"kw_only": True} if sys.version_info >= (3, 10) else {}
+
+
+@dataclass(frozen=True, **_KW_ONLY)
 class Result:
     "Concrete class for the result of a Command."
 
+    exit_code: int
+    "Command's exit code."
+
     output_bytes: Optional[bytes]
     "Output of command as bytes. May be None if there is no output."
 
-    exit_code: int
-    "Command's exit code."
+    error_bytes: bytes
+    "Limited standard error from command if not redirected."
 
     cancelled: bool
     "Command was cancelled."
@@ -42,6 +53,13 @@ class Result:
             raise TypeError("use output_bytes instead; encoding is None")
         return decode(self.output_bytes, self.encoding)
 
+    @property
+    def error(self) -> str:
+        "Error from command as a string (if it is not redirected)."
+        if self.encoding is None:
+            raise TypeError("use error_bytes instead; encoding is None")
+        return decode(self.error_bytes, self.encoding)
+
     def __bool__(self) -> bool:
         "Return true if exit_code is 0."
         return self.exit_code == 0
@@ -55,25 +73,31 @@ class PipeResult:
     cancelled: bool
 
     @staticmethod
-    def from_result(result):
+    def from_result(result: Union[Exception, Result]):
         "Construct a `PipeResult` from a `Result`."
         if isinstance(result, ResultError):
             result = result.result
+        assert isinstance(result, Result)
         return PipeResult(result.exit_code, result.cancelled)
 
 
-def make_result(command, result, cancelled, timed_out=False):
+def make_result(
+    command: "shellous.Command",
+    result_list: Union[Result, list[Union[Exception, Result]]],
+    cancelled: bool,
+    timed_out: bool = False,
+):
     """Convert list of results into a single pipe result.
 
     `result` can be a list of Result, ResultError or another Exception.
     """
-    assert result is not None
+    assert result_list is not None
 
-    if isinstance(result, list):
+    if isinstance(result_list, list):
         # Check result list for other exceptions.
         other_ex = [
             ex
-            for ex in result
+            for ex in result_list
             if isinstance(ex, Exception) and not isinstance(ex, ResultError)
         ]
         if other_ex:
@@ -81,16 +105,20 @@ def make_result(command, result, cancelled, timed_out=False):
             raise other_ex[0]
 
         # Everything in result is now either a Result or ResultError.
-        key_result = _find_key_result(result)
-        last = _get_result(result[-1])
+        key_result = _find_key_result(result_list)
+        last = _get_result(result_list[-1])
+        assert key_result is not None  # (pyright)
 
         result = Result(
-            last.output_bytes,
-            key_result.exit_code,
-            cancelled,
-            last.encoding,
-            tuple(PipeResult.from_result(r) for r in result),
+            exit_code=key_result.exit_code,
+            output_bytes=last.output_bytes,
+            error_bytes=last.error_bytes,
+            cancelled=cancelled,
+            encoding=last.encoding,
+            extra=tuple(PipeResult.from_result(r) for r in result_list),
         )
+    else:
+        result = result_list
 
     assert isinstance(result, Result)
 
@@ -110,9 +138,10 @@ def make_result(command, result, cancelled, timed_out=False):
     return result.output
 
 
-def _find_key_result(result_list):
+def _find_key_result(result_list: list[Union[Result, Exception]]) -> Result:
     "Scan a result list and return the 'key' result."
     acc = None
+
     for item in result_list:
         item = _get_result(item)
         acc = _compare_result(acc, item)
@@ -120,10 +149,13 @@ def _find_key_result(result_list):
         if (not acc.cancelled, acc.exit_code != 0) == (True, True):
             return acc
 
+    if acc is None:
+        raise RuntimeError(f"empty result: {result_list!r}")
+
     return acc
 
 
-def _compare_result(acc, item):
+def _compare_result(acc: Optional[Result], item: Result) -> Result:
     if acc is None:
         return item
     acc_key = (not acc.cancelled, acc.exit_code != 0)
@@ -133,7 +165,8 @@ def _compare_result(acc, item):
     return acc
 
 
-def _get_result(item):
+def _get_result(item: Union[Result, Exception]) -> Result:
     if isinstance(item, ResultError):
         return item.result
+    assert isinstance(item, Result)
     return item
