@@ -4,8 +4,10 @@ import asyncio
 import io
 import os
 import sys
-import warnings
 from logging import Logger
+from pathlib import Path
+from types import TracebackType
+from typing import Any, Optional, Union, cast
 
 import shellous
 import shellous.redirect as redir
@@ -13,16 +15,22 @@ from shellous import pty_util
 from shellous.harvest import harvest, harvest_results
 from shellous.log import LOG_DETAIL, LOG_ENTER, LOG_EXIT, LOGGER, log_method, log_timer
 from shellous.redirect import Redirect
-from shellous.result import Result, make_result
-from shellous.util import close_fds, poll_wait_pid, uninterrupted, verify_dev_fd, which
+from shellous.result import RESULT_STDERR_LIMIT, Result, make_result
+from shellous.util import (
+    BSD_DERIVED,
+    SupportsClose,
+    close_fds,
+    poll_wait_pid,
+    uninterrupted,
+    verify_dev_fd,
+    which,
+)
 
 _KILL_TIMEOUT = 3.0
 _CLOSE_TIMEOUT = 0.25
 _UNKNOWN_EXIT_CODE = 255
 _UNLAUNCHED_EXIT_CODE = -255
 
-_BSD = sys.platform.startswith("freebsd") or sys.platform == "darwin"
-
 AUDIT_EVENT_SUBPROCESS_SPAWN = "byllyfish/shellous.subprocess_spawn"
 """Audit event raised by sys.audit() when shellous runs a subprocess.
 
@@ -30,15 +38,11 @@ The audit event has one argument: the name of the command.
 """
 
 
-def _is_cancelled(ex):
+def _is_cancelled(ex: BaseException):
     return isinstance(ex, asyncio.CancelledError)
 
 
-def _is_cmd(cmd):
-    return isinstance(cmd, (shellous.Command, shellous.Pipeline))
-
-
-def _is_writable(cmd):
+def _is_writable(cmd: "Union[shellous.Command, shellous.Pipeline]"):
     "Return true if command/pipeline has `writable` set."
     if isinstance(cmd, shellous.Pipeline):
         # Pipelines need to check both the last/first commands.
@@ -46,7 +50,7 @@ def _is_writable(cmd):
     return cmd.options.writable
 
 
-def _enc(encoding):
+def _enc(encoding: Optional[str]):
     "Helper function to split the encoding name into codec and errors."
     if encoding is None:
         raise TypeError("when encoding is None, input must be bytes")
@@ -61,40 +65,67 @@ class _RunOptions:
 
     ```
     with _RunOptions(cmd) as options:
-        proc = await create_subprocess_exec(*options.args, **options.kwd_args)
+        proc = await create_subprocess_exec(*options.pos_args, **options.kwd_args)
         # etc.
     ```
     """
 
-    def __init__(self, command):
+    command: "shellous.Command"
+    encoding: Optional[str]
+    open_fds: list[Union[int, SupportsClose]]
+    input_bytes: Optional[bytes]
+    pos_args: list[Union[str, bytes]]
+    kwd_args: dict[str, Any]
+    subcmds: "list[Union[shellous.Command, shellous.Pipeline]]"
+    pty_fds: Optional[pty_util.PtyFds]
+    error_bytes: Optional[bytearray]
+
+    def __init__(self, command: "shellous.Command"):
         self.command = command
         self.encoding = command.options.encoding
         self.open_fds = []
         self.input_bytes = None
-        self.args = None
-        self.kwd_args = None
+        self.pos_args = []
+        self.kwd_args = {}
         self.subcmds = []
         self.pty_fds = None
-
-    def close_fds(self):
-        "Close all open file descriptors in `open_fds`."
-        close_fds(self.open_fds)
+        self.error_bytes = None
 
     def __enter__(self):
         "Set up I/O redirections."
+        _cmd = self.command
+
         try:
-            self._setup_proc_sub()
+            if _uses_process_substitution(self.command):
+                self.pos_args = self._setup_process_substitution()
+            else:
+                self.pos_args = cast(list[Union[str, bytes]], list(self.command.args))
+
             self._setup_redirects()
-            return self
+            self._setup_pass_fds()
+
+            # If executable does not include an absolute/relative directory,
+            # resolve it using PATH.
+            if not os.path.dirname(self.pos_args[0]):
+                self.pos_args[0] = which(self.pos_args[0])
+
         except Exception as ex:
             if LOG_DETAIL:
                 LOGGER.debug("_RunOptions.enter %r ex=%r", self.command.name, ex)
             _cleanup(self.command)
             raise
 
-    def __exit__(self, _exc_type, exc_value, _exc_tb):
+        assert _cmd is self.command, "self.command should remain unchanged"
+        return self
+
+    def __exit__(
+        self,
+        _exc_type: Optional[type[BaseException]],
+        exc_value: Optional[BaseException],
+        _exc_tb: Optional[TracebackType],
+    ):
         "Make sure those file descriptors are cleaned up."
-        self.close_fds()
+        close_fds(self.open_fds)
         if exc_value:
             if LOG_DETAIL:
                 LOGGER.debug(
@@ -105,42 +136,44 @@ class _RunOptions:
             for subcmd in self.subcmds:
                 _cleanup(subcmd)
 
-    def _setup_proc_sub(self):
-        "Set up process substitution."
-        if not any(_is_cmd(arg) for arg in self.command.args):
-            return
+    def _setup_process_substitution(self) -> list[Union[str, bytes]]:
+        """Set up process substitution.
 
+        Returns new command line arguments with /dev/fd path substitutions.
+        """
         if sys.platform == "win32":
             raise RuntimeError("process substitution not supported on Windows")
 
-        new_args = []
-        pass_fds = []
+        new_args: list[Union[str, bytes]] = []
+        new_fds: list[int] = []
 
         for arg in self.command.args:
-            if not _is_cmd(arg):
+            if not isinstance(arg, (shellous.Command, shellous.Pipeline)):
                 new_args.append(arg)
                 continue
 
             (read_fd, write_fd) = os.pipe()
             if _is_writable(arg):
                 new_args.append(f"/dev/fd/{write_fd}")
-                pass_fds.append(write_fd)
+                new_fds.append(write_fd)
                 subcmd = arg.stdin(read_fd, close=True)
             else:
                 new_args.append(f"/dev/fd/{read_fd}")
-                pass_fds.append(read_fd)
+                new_fds.append(read_fd)
                 subcmd = arg.stdout(write_fd, close=True)
 
             self.subcmds.append(subcmd)
 
-        # pylint: disable=protected-access
-        self.command = self.command._replace_args(new_args).set(
-            pass_fds=pass_fds,
-            pass_fds_close=True,
-        )
+        # We need to include `new_fds` in our `pass_fds` list. We also need
+        # to close them after process launch.
+        self.kwd_args["pass_fds"] = new_fds
+        self.kwd_args["close_fds"] = True
+        self.open_fds.extend(new_fds)
 
-        if _BSD:
-            verify_dev_fd(pass_fds[0])
+        if BSD_DERIVED:
+            verify_dev_fd(new_fds[0])
+
+        return new_args
 
     def _setup_redirects(self):
         "Set up I/O redirections."
@@ -183,25 +216,29 @@ class _RunOptions:
         assert not preexec_fn or callable(preexec_fn)
 
         self.input_bytes = input_bytes
-        self.kwd_args = {
-            "stdin": stdin,
-            "stdout": stdout,
-            "stderr": stderr,
-            "env": options.merge_env(),
-            "start_new_session": start_session,
-            "preexec_fn": preexec_fn,
-            "close_fds": options.close_fds,
-        }
+        self.kwd_args.update(
+            stdin=stdin,
+            stdout=stdout,
+            stderr=stderr,
+            env=options.merge_env(),
+            start_new_session=start_session,
+            preexec_fn=preexec_fn,
+        )
+
+    def _setup_pass_fds(self):
+        "Set up `pass_fds` and `close_fds` if pass_fds is configured."
+        options = self.command.options
 
         if options.pass_fds:
+            # Setting pass_fds causes close_fds to be True. Set close_fds to
+            # True manually to avoid a RuntimeWarning.
             self.kwd_args["close_fds"] = True
-            self.kwd_args["pass_fds"] = options.pass_fds
+            self.kwd_args.setdefault("pass_fds", []).extend(options.pass_fds)
             if options.pass_fds_close:
                 self.open_fds.extend(options.pass_fds)
-
-        self.args = list(self.command.args)
-        if not os.path.dirname(self.args[0]):
-            self.args[0] = which(self.args[0])
+        elif "close_fds" not in self.kwd_args:
+            # Only set close_fds if it is not already set.
+            self.kwd_args["close_fds"] = options.close_fds
 
     def _setup_input(self, input_, close, encoding):
         "Set up process input."
@@ -212,13 +249,15 @@ class _RunOptions:
 
         if isinstance(input_, (bytes, bytearray)):
             input_bytes = input_
-        elif isinstance(input_, os.PathLike):
+        elif isinstance(input_, Path):
             stdin = open(input_, "rb")  # pylint: disable=consider-using-with
             self.open_fds.append(stdin)
         elif isinstance(input_, Redirect) and input_.is_custom():
             # Custom support for Redirect constants.
             if input_ == Redirect.INHERIT:
                 stdin = sys.stdin
+            elif input_ == Redirect.RESULT:
+                raise TypeError(f"unsupported input type: {input_!r}")
             else:
                 # CAPTURE uses stdin == PIPE.
                 assert input_ == Redirect.CAPTURE
@@ -244,19 +283,21 @@ class _RunOptions:
 
         stdout = asyncio.subprocess.PIPE
 
-        # TODO: str|bytes as output are deprecated.
         if isinstance(output, (str, bytes, os.PathLike)):
-            if isinstance(output, (str, bytes)):
-                warnings.warn(
-                    "Using str|bytes as output file is deprecated",
-                    DeprecationWarning,
-                )
+            if isinstance(output, (str, bytes)):  # TODO: remove this check...
+                raise TypeError("Using str|bytes as output file is unsupported")
             mode = "ab" if append else "wb"
             stdout = open(output, mode=mode)  # pylint: disable=consider-using-with
             self.open_fds.append(stdout)
         elif isinstance(output, Redirect) and output.is_custom():
             # Custom support for Redirect constants.
-            if output == Redirect.INHERIT:
+            if output == Redirect.RESULT:
+                # We do not support redirecting stdout to RESULT (only stderr).
+                if sys_stream == sys.stdout:
+                    raise TypeError(f"unsupported output type: {output!r}")
+                assert stdout == asyncio.subprocess.PIPE
+                self.error_bytes = bytearray()
+            elif output == Redirect.INHERIT:
                 stdout = sys_stream
             else:
                 # CAPTURE uses stdout == PIPE.
@@ -287,18 +328,19 @@ class _RunOptions:
         """
 
         self.pty_fds = pty_util.open_pty(pty)
-        child_fd = int(self.pty_fds.child_fd)
 
         # On BSD-derived systems like FreeBSD and Darwin, we delay closing the
         # pty's child_fd in the parent process until after the first read
         # succeeds. On Linux, we close the child_fd as soon as possible.
 
-        if not _BSD:
+        if not BSD_DERIVED:
             self.open_fds.append(self.pty_fds.child_fd)
 
         if LOG_DETAIL:
             LOGGER.info("_setup_pty1: %r", self.pty_fds)
 
+        child_fd = int(self.pty_fds.child_fd)
+
         if stdin == asyncio.subprocess.PIPE:
             stdin = child_fd
 
@@ -383,12 +425,14 @@ class Runner:
         assert code is not None
 
         result = Result(
-            output_bytes,
-            code,
-            self._cancelled,
-            self._options.encoding,
+            exit_code=code,
+            output_bytes=output_bytes,
+            error_bytes=bytes(self._options.error_bytes or b""),
+            cancelled=self._cancelled,
+            encoding=self._options.encoding,
         )
 
+        # TODO: This method should return a Result object... (not str|bytes)
         return make_result(self.command, result, self._cancelled, self._timed_out)
 
     def add_task(self, coro, tag=""):
@@ -398,9 +442,19 @@ class Runner:
         self._tasks.append(task)
         return task
 
+    def send_signal(self, sig: int):
+        "Send an arbitrary signal to the process if it is running."
+        if self.returncode is None:
+            self._signal(sig)
+
+    def cancel(self):
+        "Cancel the running process if it is running."
+        if self.returncode is None:
+            self._signal(self.command.options.cancel_signal)
+
     def _is_bsd_pty(self):
         "Return true if we're running a pty on BSD."
-        return _BSD and self._options.pty_fds
+        return BSD_DERIVED and self._options.pty_fds
 
     @log_method(LOG_DETAIL)
     async def _wait(self):
@@ -429,7 +483,10 @@ class Runner:
     async def _wait_pid(self):
         "Manually poll `waitpid` until process finishes."
         assert self._is_bsd_pty()
+
         while True:
+            assert self._proc is not None  # (pyright)
+
             if poll_wait_pid(self._proc):
                 break
             await asyncio.sleep(0.025)
@@ -445,7 +502,7 @@ class Runner:
         try:
             # If not already done, send cancel signal.
             if self._proc.returncode is None:
-                self._send_signal(cancel_signal)
+                self._signal(cancel_signal)
 
             if self._tasks:
                 await harvest(*self._tasks, timeout=cancel_timeout, trustee=self)
@@ -464,8 +521,9 @@ class Runner:
             await self._kill_wait()
             raise
 
-    def _send_signal(self, sig):
+    def _signal(self, sig: Optional[int]):
         "Send a signal to the process."
+        assert self._proc is not None  # (pyright)
 
         LOGGER.info("Runner.signal %r signal=%r", self, sig)
         self._audit_callback("signal", signal=sig)
@@ -486,7 +544,7 @@ class Runner:
             return
 
         try:
-            self._send_signal(None)
+            self._signal(None)
             await harvest(self._waiter(), timeout=_KILL_TIMEOUT, trustee=self)
         except asyncio.TimeoutError as ex:
             # Manually check if the process is still running.
@@ -521,6 +579,7 @@ class Runner:
             # Set up subprocess arguments and launch subprocess.
             with self._options as opts:
                 await self._subprocess_spawn(opts)
+                assert self._proc is not None  # (pyright)
 
             stdin = self._proc.stdin
             stdout = self._proc.stdout
@@ -532,8 +591,15 @@ class Runner:
                 stdin, stdout = opts.pty_fds.writer, opts.pty_fds.reader
 
             if stderr is not None:
-                error = opts.command.options.error
-                stderr = self._setup_output_sink(stderr, error, opts.encoding, "stderr")
+                limit = -1
+                if opts.error_bytes is not None:
+                    error = opts.error_bytes
+                    limit = RESULT_STDERR_LIMIT
+                else:
+                    error = opts.command.options.error
+                stderr = self._setup_output_sink(
+                    stderr, error, opts.encoding, "stderr", limit
+                )
 
             if stdout is not None:
                 output = opts.command.options.output
@@ -589,19 +655,23 @@ class Runner:
             self.add_task(cmd.coro(), "procsub")
 
     @log_method(LOG_DETAIL)
-    async def _subprocess_exec(self, opts):
+    async def _subprocess_exec(self, opts: _RunOptions):
         "Start the subprocess and assign to `self.proc`."
         with log_timer("asyncio.create_subprocess_exec"):
-            sys.audit(AUDIT_EVENT_SUBPROCESS_SPAWN, opts.args[0])
-            with pty_util.set_ignore_child_watcher(opts.pty_fds and _BSD):
+            sys.audit(AUDIT_EVENT_SUBPROCESS_SPAWN, opts.pos_args[0])
+            with pty_util.set_ignore_child_watcher(
+                BSD_DERIVED and opts.pty_fds is not None
+            ):
                 self._proc = await asyncio.create_subprocess_exec(
-                    *opts.args,
+                    *opts.pos_args,
                     **opts.kwd_args,
                 )
 
     @log_method(LOG_DETAIL)
     async def _waiter(self):
         "Run task that waits for process to exit."
+        assert self._proc is not None  # (pyright)
+
         try:
             if self._is_bsd_pty():
                 await self._wait_pid()
@@ -663,7 +733,7 @@ class Runner:
 
         return stream
 
-    def _setup_output_sink(self, stream, sink, encoding, tag):
+    def _setup_output_sink(self, stream, sink, encoding, tag, limit=-1):
         "Set up a task to write to custom output sink."
         if isinstance(sink, io.StringIO):
             if encoding is None:
@@ -674,7 +744,11 @@ class Runner:
             self.add_task(redir.copy_bytesio(stream, sink), tag)
             stream = None
         elif isinstance(sink, bytearray):
-            self.add_task(redir.copy_bytearray(stream, sink), tag)
+            # N.B. `limit` is only supported for bytearray.
+            if limit >= 0:
+                self.add_task(redir.copy_bytearray_limit(stream, sink, limit), tag)
+            else:
+                self.add_task(redir.copy_bytearray(stream, sink), tag)
             stream = None
         elif isinstance(sink, Logger):
             self.add_task(redir.copy_logger(stream, sink, encoding), tag)
@@ -726,23 +800,21 @@ class Runner:
     @log_method(LOG_DETAIL)
     async def _close(self):
         "Make sure that our resources are properly closed."
-        assert self._proc
+        assert self._proc is not None
 
         if self._options.pty_fds:
             self._options.pty_fds.close()
 
+        # Make sure the transport is closed (for asyncio and uvloop).
+        self._proc._transport.close()  # pyright: ignore pylint: disable=protected-access
+
         # _close can be called when unwinding exceptions. We need to handle
-        # the case that the process has not exited yet. Remember to close the
-        # transport.
+        # the case that the process has not exited yet.
         if self._proc.returncode is None:
             LOGGER.critical("Runner._close process still running %r", self._proc)
-            self._proc._transport.close()  # pylint: disable=protected-access
             return
 
         try:
-            # Make sure the transport is closed (for asyncio and uvloop).
-            self._proc._transport.close()  # pylint: disable=protected-access
-
             # Make sure that original stdin is properly closed. `wait_closed`
             # will raise a BrokenPipeError if not all input was properly written.
             if self._proc.stdin is not None:
@@ -761,7 +833,7 @@ class Runner:
         "Call `audit_callback` if there is one."
         callback = self.command.options.audit_callback
         if callback:
-            info = {"runner": self}
+            info: dict[str, Any] = {"runner": self}
             if failure:
                 info["failure"] = failure
             if phase == "signal":
@@ -792,7 +864,11 @@ class Runner:
         return self._readlines()
 
     @staticmethod
-    async def run_command(command, *, _run_future=None):
+    async def run_command(
+        command: "shellous.Command",
+        *,
+        _run_future: Optional[asyncio.Future["Runner"]] = None,
+    ):
         "Run a command. This is the main entry point for Runner."
         if not _run_future and _is_multiple_capture(command):
             LOGGER.warning("run_command: multiple capture requires 'async with'")
@@ -1018,7 +1094,7 @@ class PipeRunner:
         return self._readlines()
 
     @staticmethod
-    async def run_pipeline(pipe):
+    async def run_pipeline(pipe: "shellous.Pipeline"):
         "Run a pipeline. This is the main entry point for PipeRunner."
 
         run = PipeRunner(pipe, capturing=False)
@@ -1027,6 +1103,13 @@ class PipeRunner:
         return run.result()
 
 
+def _uses_process_substitution(cmd: "shellous.Command") -> bool:
+    "Return True if command uses process substitution."
+    return any(
+        isinstance(arg, (shellous.Command, shellous.Pipeline)) for arg in cmd.args
+    )
+
+
 def _is_multiple_capture(cmd):
     "Return true if both stdout and stderr are CAPTURE."
     output = Redirect.from_default(cmd.options.output, 1, cmd.options.pty)
