@@ -2,17 +2,32 @@
 
 import asyncio
 import contextvars
-import io
 import os
 import shutil
+import sys
 from asyncio.subprocess import Process
-from collections import defaultdict
-from typing import Any, Iterable, Optional, Union
+from collections import abc, defaultdict
+from types import TracebackType
+from typing import (
+    Any,
+    AsyncContextManager,
+    Coroutine,
+    Iterable,
+    Optional,
+    Protocol,
+    Union,
+)
 
 from .log import LOG_DETAIL, LOGGER, log_timer
 
 # Stores current stack of context managers for immutable Command objects.
-_CTXT_STACK = contextvars.ContextVar("ctxt_stack", default=None)
+_CTXT_STACK = contextvars.ContextVar[Optional[dict[int, list[Any]]]](
+    "ctxt_stack",
+    default=None,
+)
+
+# True if OS is derived from BSD.
+BSD_DERIVED = sys.platform.startswith("freebsd") or sys.platform == "darwin"
 
 
 def decode(data: Optional[bytes], encoding: str) -> str:
@@ -29,7 +44,7 @@ def coerce_env(env: dict[str, Any]) -> dict[str, str]:
     parent environment.
     """
 
-    def _coerce(key, value):
+    def _coerce(key: str, value: Any):
         if value is ...:
             value = os.environ[key]
         return str(value)
@@ -37,7 +52,14 @@ def coerce_env(env: dict[str, Any]) -> dict[str, str]:
     return {str(key): _coerce(key, value) for key, value in env.items()}
 
 
-def close_fds(open_fds: Iterable[Union[io.IOBase, int]]) -> None:
+class SupportsClose(Protocol):
+    "Protocol for objects with a close() method."
+
+    def close(self) -> None:
+        ...
+
+
+def close_fds(open_fds: Iterable[Union[SupportsClose, int]]) -> None:
     "Close open file descriptors or file objects."
     with log_timer("close_fds"):
         try:
@@ -47,7 +69,7 @@ def close_fds(open_fds: Iterable[Union[io.IOBase, int]]) -> None:
                         try:
                             os.close(obj)
                         except OSError as ex:
-                            LOGGER.warning("os.close ex=%r", ex)
+                            LOGGER.warning("os.close ex=%r", ex, stack_info=True)
                 else:
                     obj.close()
         finally:
@@ -119,7 +141,7 @@ def poll_wait_pid(proc: Process) -> bool:
     return True
 
 
-async def uninterrupted(coro):
+async def uninterrupted(coro: Coroutine[Any, Any, Any]):
     "Run a coroutine so it completes even if the current task is cancelled."
 
     task = asyncio.create_task(coro)
@@ -132,7 +154,7 @@ async def uninterrupted(coro):
         raise
 
 
-def which(command):
+def which(command: str):
     "Given a command without a directory, return the fully qualified path."
     path = shutil.which(command)
     if path is None:
@@ -140,23 +162,29 @@ def which(command):
     return path
 
 
-async def context_aenter(scope, ctxt_manager):
+async def context_aenter(scope: int, ctxt_manager: AsyncContextManager[Any]):
     "Enter an async context manager."
     ctxt_stack = _CTXT_STACK.get()
     if ctxt_stack is None:
-        ctxt_stack = defaultdict(list)
+        ctxt_stack = defaultdict[int, list[Any]](list)
         _CTXT_STACK.set(ctxt_stack)
 
-    result = await ctxt_manager.__aenter__()
+    result = await ctxt_manager.__aenter__()  # pylint: disable=unnecessary-dunder-call
     stack = ctxt_stack[scope]
     stack.append(ctxt_manager)
 
     return result
 
 
-async def context_aexit(scope, exc_type, exc_value, exc_tb):
+async def context_aexit(
+    scope: int,
+    exc_type: Optional[type[BaseException]],
+    exc_value: Optional[BaseException],
+    exc_tb: Optional[TracebackType],
+):
     "Exit an async context manager."
     ctxt_stack = _CTXT_STACK.get()
+    assert ctxt_stack is not None  # (pyright)
 
     stack = ctxt_stack[scope]
     ctxt_manager = stack.pop()
@@ -166,3 +194,38 @@ async def context_aexit(scope, exc_type, exc_value, exc_tb):
         _CTXT_STACK.set(None)
 
     return await ctxt_manager.__aexit__(exc_type, exc_value, exc_tb)
+
+
+class EnvironmentDict(abc.Mapping[str, str]):
+    "Read-only, hashable dictionary that stores environment variables."
+
+    _data: dict[str, str]
+
+    def __init__(self, base: Optional["EnvironmentDict"], updates: dict[str, Any]):
+        if base is None:
+            self._data = {}
+        else:
+            self._data = base._data.copy()
+        self._data.update(**coerce_env(updates))
+
+    def __getitem__(self, key: str) -> str:
+        return self._data[key]
+
+    def __len__(self) -> int:
+        return len(self._data)
+
+    def __iter__(self):
+        return iter(self._data)
+
+    def __hash__(self) -> int:
+        return hash(tuple((key, value) for key, value in self._data.items()))
+
+    def __eq__(self, rhs: object) -> bool:
+        if isinstance(rhs, dict):
+            return self._data == rhs
+        if not isinstance(rhs, EnvironmentDict):
+            return False
+        return self._data == rhs._data
+
+    def __repr__(self) -> str:
+        return repr(self._data)
